static char *rcsid = "$Id: mpx.c,v 1.2 1996/07/09 19:41:39 paul Exp $";

/*
 * This file contains functions to handle context switching and SunOS signals
 * in smx.  As signals cause context switches, the two are closely related.
 * The entry points for the file are:
 *
 *       restart - called from the bootstrap code to schedule the first process
 *       idle_task - the main body of the idle task.
 *
 * The SunOSsig function could also be considered an entry point.  It is
 * the handler for all SunOS signals, and so is called whenever a SunOS
 * signal occurs.  It is the sole entry point for the layer 1 code once
 * the first process has been scheduled for the first time.
 */

#include "kernel.h"
#include "proc.h"
#include "sunsighandle.h"

#include <sun/syscall.h>

/*
 * Check that the context size calculated by make_context (UC_SIZE)
 * matches the smx CONTEXT_SIZE.
 */
#if (CONTEXT_SIZE != UC_SIZE)
error "p_reg is the wrong size"
#endif

static double gwin_save_area[GWIN_SIZE / sizeof(double) + 1];
struct proc *gwin_proc = 0;


/*
 * mpx_start and mpx_end are dummy functions used to delimit the section
 * of this file (the resume and SunOSsig functions) that must remain
 * executable even when execution is outside the kernel.  These labels
 * are needed by code in sunprotect.c
 */

void mpx_start() {}

/*
 * Function: resume
 * Parameter: ctxt - double-word aligned context area that we can copy
 *                   the context of proc_ptr.
 *
 * Resume execution of the the current process as given by proc_ptr.
 * We make a local copy of the context to restore for two reasons: to double
 * word align it, and so that if full memory protection is enabled we
 * still have access to the context after leaving_kernel has done its work.
 * Callers must ensure that ctxt points to a double-word aligned area on the
 * stack.  If a gwin structure is to be restored, it is first copied to
 * an array local to resume for the same reasons.
 */
static void resume(reg_t *ctxt)
{
    double local_save_area[GWIN_SIZE / sizeof(double) + 1];
    
    memcpy(ctxt, &proc_ptr->p_reg, sizeof(proc_ptr->p_reg));
    
    /*
     * Clear signal mask---no SunOS signals are masked when execution is
     * outside layer 1.
     */
    memset(&ctxt[UC_SIGMASK], 0, SIGSET_SIZE);

    /*
     * Set the sig mask flag bit to ensure the signal mask gets restored.
     */
    ctxt[UC_FLAGS] |= 1;
    if (ctxt[UC_GWINS] != 0) {
        /*
	 * A gwins structure must be restored.  Copy it in, and update the
         * gwin pointer in the context to be restored.
	 */
	gwin_proc = 0;
	memcpy(local_save_area, gwin_save_area, sizeof(local_save_area));
	ctxt[UC_GWINS] = (reg_t) local_save_area;
    }
    leaving_kernel();

    /*
     * Do a context switch.  This call should never return, with layer 1 code
     * being re-entered via SunOSsig.
     */
    SunOS(SYS_context, 1, ctxt);
    printf("Gwin = 0x%x, fpuq = 0x%x, ctxt = 0x%x\n", ctxt[UC_GWINS],
	   ctxt[UC_FPU_Q], ctxt);
    panic("resume: failed to resume an smx process", errno);
}    


/*
 * Function: SunOSsig
 * Parameters: sig - the SunOS signal that has occurred
 *             siginfo - additional info about the signal
 *             ucontext - context for the smx process that was running when
 *                        the SunOS signal ocurred.
 *
 * This function is the handler for all signals received by the SunOS
 * process running smx.  The 'ucontext' structure pushed onto the
 * kernel (signal) stack by Solaris is copied into the MINIX proc
 * table entry of the currently executing MINIX process. The array
 * 'vectors' is used to determine which function within the smx
 * kernel should then take responsibilty for the signal. Finally, the
 * context of an smx process is restored.
 */
void SunOSsig(int sig, void *siginfo, reg_t ucontext[])
{
    entering_kernel();

    /*
     * Is there a gwins structure to save (should only happen when the stack
     * of a user process needs extending).  The single save area is sufficient
     * because once it is in use the process using it will be the next user 
     * process to run (the scheduler ensures this).
     */
    if (ucontext[UC_GWINS] != 0) {
	if (gwin_proc != 0) {
	    panic("SunOSsig---save area used", NO_NUM);
	}
	if (!isuserp(proc_ptr)) {
	    panic("UC_GWINS non-zero in process", proc_number(proc_ptr));
	}
	gwin_proc = proc_ptr;
	memcpy(gwin_save_area, (char *) ucontext[UC_GWINS],
	       sizeof(gwin_save_area));
    }
    /*
     * Save the ucontext structure into the current proc table entry.
     */
     memcpy(&proc_ptr->p_reg, ucontext, sizeof(proc_ptr->p_reg));


    if (sig < 0 || sig >= vectorlen)
	panic("SunOS signal too large or too small", sig);
    (*vectors[sig])(sig, siginfo, ucontext);      /* Call the handler */
	
    lock_pick_proc();         /* (possibly) choose a new process */

    resume(ucontext);
    /*NOTREACHED*/
}

void mpx_end() {}


/*
 * Function: restart
 *
 * The 'restart' function is called (only once) from 'main.c' in order
 * to get the the first process running. 
 */
void restart(void)
{
    struct stackframe_s dummy_regs;

    k_reenter = 0;      /* Nested interrupt handling is not supported */

    resume((reg_t *) &dummy_regs);
    /*NOTREACHED*/
}


/*****************************************************************************
 * Some of the functions found in the vectors array.                         *
 *****************************************************************************/

/*
 * Function: s_call
 *
 * s_call is called from SunOSsig to handle the SunOS USR1 signal,
 * which is sent by an smx process in layers 2-4 to force execution
 * into layer 1 of smx.  The three system calls are send, receive,
 * and sndrcv.  s_call simply sets up parameters for a call to the
 * standard MINIX sys_call function, and passes the return result
 * back to the calling smx process.
 */
void s_call(void)
{
    phys_bytes param_ptr;       /* ptr to the system call parameters */
    int function;		/* SEND, RECEIVE, or BOTH */
    int src_dest;		/* source to receive from or dest to send to */
    message *m_ptr;		/* pointer to message */
    int params[3];
    struct proc *this_proc = proc_ptr;

    /*
     * The arguments to sys_call are on the stack.  For the SPARC
     * stack frame, the first argument is at sp + 92.
     */
    param_ptr = umap(this_proc, D, this_proc->p_reg.sp + 92, sizeof(params));
    if (param_ptr == 0) {
	panic("s_call: Can't map in parameter to system call", NO_NUM);
    }

    /* Load the parameters */
    phys_copy(param_ptr, vir2phys(&params[0]), sizeof(params));
    function = params[0];
    src_dest = params[1];
    m_ptr = (message *)params[2];   /* Assumes ints can hold a pointer */

    this_proc->p_reg.retreg = sys_call(function, src_dest, m_ptr);
}


/*
 * Function: smx_clock_handler
 *
 * Called by SunOSsig whenever a SIGALRM occurs.  Calls clock handler
 * with 0 ticks, as in smx the SunOS clock is always consulted to
 * update the time on a clock tick.
 */
void smx_clock_handler(void)
{
    clock_handler(0);
}


/*==========================================================================*
 *                             idle_task                                    * 
 *==========================================================================*/

#define SIGPAUSE 0x1000

/* Does nothing but sleep waiting for a signal*/

void idle_task(void)
{
    for (;;) {
	SunOS(SYS_signal, (SIGPAUSE | 1));
    }
}
