static char *rcsid = "$Id$";

/*
 * This file contains functions that must work closely with the SunOS
 * signal handling functions.  These functions relate to initialising
 * the handling of SunOS signals, and implementing lock and unlock.
 * A special feature of this file is that we include the SunOS signal.h
 * for use by the functions in the file.  To achieve this, the only smx
 * header files included are those that define constants---type clashes
 * must be avoided.
 *
 * The entry points for the file are:
 *
 *       init_sun_handlers - installs handlers for all SunOS signals
 *       lock - disable non-error SunOS signals
 *       unlock - restore the signal mask to its state prior to the lock call
 */

#include <minix/config.h>
#include <minix/const.h>
#include "const.h"
#include "sunsighandle.h"

#include <sun/signal.h>
#include <sun/syscall.h>

#define __EXTENSIONS__
#include </usr/include/sys/types.h>     /* Get the sys/types.h we want */
#include </usr/include/signal.h>

/*
 * Local functions.
 */
static void get_inkernel_mask(so_sigset_t *sigset);

/*
 * vectors specifies for each SunOS signal the kernel function that
 * should be called to handle it.  It is defined and initialised in this 
 * file, and used in mpx.c.
 */ 
void (*vectors[_SIGRTMAX + 1])();
int vectorlen = sizeof(vectors) / sizeof(vectors[0]);


/*
 * Function: init_sun_handlers
 *
 * Sets up the handling of all SunOS signals by:
 *    - installing the kernel stack as the alternate signal stack so that
 *      all signal handling will make use of that stack.
 *    - installing SunOSsig as the handler for all functions
 *    - setting up vectors to record the kernel function to call on the
 *      occurrence of each smx signal.
 */
void init_sun_handlers(void)
{
    struct sigaction sig; 
    stack_t sstk;
    int t;

    if (sizeof(sigset_t) != sizeof(so_sigset_t)) {
	panic("so_sigset_t is the wrong size", NO_NUM);
    }

    /*
     * tell SunOS the location of the stack to use for handling 
     * SunOS signals
     */
    sstk.ss_sp = (char *)getksp() - K_STACK_BYTES;
    sstk.ss_size = K_STACK_BYTES; 
    sstk.ss_flags = 0;
    SunOS(SYS_sigaltstack, &sstk, (struct sigaltstack *)0);
    debug_str("Kernel stack from ");
    debug_int((int) sstk.ss_sp); debug_str("\n");

    /* 
     * each of SunOS's signals go through a special handler that takes the
     * current Minix process's context and stores it in the proc table 
     */ 
    sig.sa_flags = SA_ONSTACK | SA_SIGINFO;
    sig.sa_handler = SunOSsig; 
    get_inkernel_mask((so_sigset_t *) &sig.sa_mask); 
    for (t = 1; t <= _SIGRTMAX; t++) {
 	SunOS(SYS_sigaction, t, &sig, (struct sigaction *) 0); 
    }

    /* 
     * The SunOSsig signal handler uses 'vectors' to decide which function
     * should be called to handle each SunOS signal (eg. SIGIO).
     */
    for (t = 0; t <= _SIGRTMAX; t++) { 	
	vectors[t] = exception; 
    }

    vectors[SIGUSR1] = s_call; 
    vectors[SIGIO] = sunio_interrupt;
    vectors[SIGALRM] = smx_clock_handler; 
}


/*==========================================================================*
 *                             lock etc                                     * 
 *==========================================================================*/

/*
 * It is assumed that lock/unlock calls are made in pairs.  Initially
 * the system is not locked (lockcnt = 0).  When lock is called, it
 * sets the SunOS signal mask so as to block all non-error signals,
 * and increments lockcnt.  When unlock is called it checks lockcnt.
 * If lockcnt is 0 or less then a fatal error has occurred, as the
 * lock operation has not been done before the unlock.  If lockcnt is one
 * then the SunOS signal mask is set to an empty mask (allowing all signals)
 * and lockcnt is set to 0.  If lockcnt is greater than one then the signal
 * mask is unchanged, and lockcnt is reduced by 1,
 */

static int lockcnt = 0;      /* Used to enforce lock/unlock pairs */


/*
 * Function: get_inkernel_mask
 * Returns: the SunOS signal mask that should be installed while execution
 *          is in layer 1.  All but "error" signals should be masked out.
 */
static void get_inkernel_mask(so_sigset_t *mask)
{
    static so_sigset_t lockmask;
    static int called_before = 0;

    if (!called_before) {
	/*
	 * Initialise lockmask.  Only Sun error signals should be allowed
	 * through.  First all signals are masked with the call to memset,
	 * then the error signals are allowed through.
	 */
	sunsigfullset(&lockmask);
	sunsigdelset(&lockmask, SIGILL);
	sunsigdelset(&lockmask, SIGTRAP);
	sunsigdelset(&lockmask, SIGEMT);
	sunsigdelset(&lockmask, SIGFPE);
	sunsigdelset(&lockmask, SIGBUS);
	sunsigdelset(&lockmask, SIGSEGV);
	sunsigdelset(&lockmask, SIGSYS); 
	called_before = 1;
    }
    *mask = lockmask;
}


/*
 * Function: lock
 *
 * Sets the signal mask to allow delivery of "error" signals only.
 * Lockcnt is used to remember the number of outstanding lock calls.
 */
void lock()
{
    so_sigset_t lockmask;    /* Signal mask to use in locked periods */

    get_inkernel_mask(&lockmask);
    SunOS(SYS_sigprocmask, SIG_SETMASK, &lockmask, 0);
    lockcnt++;
}


/*
 * Function: unlock
 *
 * If this undoes the last lock, then a mask is installed to allow
 * delivery of all SunOS signals.  Otherwise, lockcnt is decremented
 * to record one less outstanding lock call.
 */
void unlock(void) 
{
    so_sigset_t clearmask;

    if (lockcnt <= 0) {
	panic("unlock() called without a prior call to lock()", NO_NUM);
    }
    if (lockcnt == 1) {
	lockcnt = 0;
	sunsigemptyset(&clearmask);
	SunOS(SYS_sigprocmask, SIG_SETMASK, &clearmask, 0);
    } else {
	lockcnt--;
    }
}
